/*
 *  ep93xx_gpio.c
 *
 *  Copyright (C) 2004-2005 Atmark Techno, Inc.
 */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/ep93xx_gpio.h>
#include <asm/hardware.h>
#include <asm/io.h>

#ifndef GPIO_MINOR
#define GPIO_MINOR 186
#endif

static const __u32 GPIO_REG[EP93XX_GPIO_NUM] = {
  GPIO_PADR,
  GPIO_PBDR,
  GPIO_PCDR,
  GPIO_PDDR,
  GPIO_PADDR,
  GPIO_PBDDR,
  GPIO_PCDDR,
  GPIO_PDDDR,
  GPIO_PEDR,
  GPIO_PEDDR,
  GPIO_PFDR,
  GPIO_PFDDR,
  GPIO_PGDR,
  GPIO_PGDDR,
  GPIO_PHDR,
  GPIO_PHDDR,
};

int ep93xx_gpio_in(__u32 device, __u32 mask, volatile __u32 *data)
{
	if (device >= EP93XX_GPIO_NUM) {
		return -EFAULT;
	}

	*data &= ~mask;
	*data |= inl(GPIO_REG[device]) & mask;

	return 0;
}

int ep93xx_gpio_out(__u32 device, __u32 mask, __u32 data)
{
	__u32 tmp = 0;

	if (device >= EP93XX_GPIO_NUM) {
		return -EFAULT;
	}

	if (~mask) {
		tmp = inl(GPIO_REG[device]) & (~mask);
	}
	outl(tmp | (data & mask), GPIO_REG[device]);

	return 0;
}

static int ep93xx_gpio_open(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev); 
	if (minor != GPIO_MINOR)
		return -ENODEV;

#ifdef MODULE
	//MOD_INC_USE_COUNT;
#endif

	return 0;
}

static int ep93xx_gpio_release(struct inode *inode, struct file *file)
{

#ifdef MODULE
	//MOD_DEC_USE_COUNT;
#endif

	return 0;
}

static int ep93xx_gpio_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{

	static struct ep93xx_gpio_ioctl_data ioctl_data;

	if (copy_from_user(&ioctl_data, (struct ep93xx_gpio_ioctl_data *)arg,
			sizeof(ioctl_data)))
		return -EFAULT;

	switch(cmd) {
		case EP93XX_GPIO_IN:
			if (ep93xx_gpio_in(ioctl_data.device, ioctl_data.mask,
					   &ioctl_data.data))
				return -EFAULT;

			if (copy_to_user((struct ep93xx_gpio_ioctl_data *)arg,
					 &ioctl_data, sizeof(ioctl_data)))
				return -EFAULT;
			break;

		case EP93XX_GPIO_OUT:
			return ep93xx_gpio_out(ioctl_data.device,
					ioctl_data.mask, ioctl_data.data);
			break;

		default:
			return -ENOIOCTLCMD;

	}
	return 0;
}

static struct file_operations ep93xx_gpio_fops =
{
	owner:		THIS_MODULE,
	ioctl:		ep93xx_gpio_ioctl,
	open:		ep93xx_gpio_open,
	release:	ep93xx_gpio_release,
};

/* GPIO_MINOR in include/linux/miscdevice.h */
static struct miscdevice ep93xx_gpio_miscdev =
{
	GPIO_MINOR,
	"ep93xx_gpio",
	&ep93xx_gpio_fops
};

int __init ep93xx_gpio_init(void)
{
	misc_register(&ep93xx_gpio_miscdev);

	if (check_mem_region(GPIO_PADR, 0x48) < 0 ) {
           return -ENODEV;
        } else {
           request_mem_region(GPIO_PADR, 0x48, "EP93XX GPIO");
        }     

	return 0;
}	

void __exit ep93xx_gpio_exit(void)
{
	misc_deregister(&ep93xx_gpio_miscdev);
}

module_init(ep93xx_gpio_init);
module_exit(ep93xx_gpio_exit);

MODULE_LICENSE("GPL v2");
